# Copyright 2012-2014 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

from tests import unittest
from botocore.translate import ModelFiles, translate, merge_dicts, \
                               resembles_jmespath_exp


SERVICES = {
  "service_full_name": "AWS Security Token Service",
  "service_abbreviation": "AWS STS",
  "type": "query",
  "signature_version": "v4",
  "result_wrapped": True,
  "global_endpoint": "sts.amazonaws.com",
  "api_version": "2011-06-15",
  "endpoint_prefix": "sts",
  "xmlnamespace": "https://sts.amazonaws.com/doc/2011-06-15/",
  "documentation": "docs",
  "operations": {
    "AssumeRole": {
      "name": "AssumeRole",
      "input": {
        "shape_name": "AssumeRoleRequest",
        "type": "structure",
        "members": {
          "RoleArn": {
            "shape_name": "arnType",
            "type": "string",
            "min_length": 20,
            "max_length": 2048,
            "documentation": "docs",
            "required": True
          },
          "RoleSessionName": {
            "shape_name": "userNameType",
            "type": "string",
            "min_length": 2,
            "max_length": 32,
            "pattern": "[\\w+=,.@-]*",
            "documentation": "docs",
            "required": True
          },
          "Policy": {
            "shape_name": "sessionPolicyDocumentType",
            "type": "string",
            "pattern": "[\\u0009\\u000A\\u000D\\u0020-\\u00FF]+",
            "min_length": 1,
            "max_length": 2048,
            "documentation": "docs"
          },
          "DurationSeconds": {
            "shape_name": "roleDurationSecondsType",
            "type": "integer",
            "min_length": 900,
            "max_length": 3600,
            "documentation": "docs"
          },
          "ExternalId": {
            "shape_name": "externalIdType",
            "type": "string",
            "min_length": 2,
            "max_length": 96,
            "pattern": "[\\w+=,.@:-]*",
            "documentation": "docs"
          },
          # Not in the actual description, but this is to test pagination.
          "NextToken": {
              "shape_name": "String",
              "type": "string",
              "documentation": None
          },
          "TokenToken": {
              "shape_name": "String",
              "type": "string",
              "documentation": None,
              "xmlname": "tokenToken"
          },
          "MaxResults": {
              "shape_name": "Integer",
              "type": "int",
              "documentation": None,
              "xmlname": "maxResults"
          }
        },
        "documentation": "docs"
      },
      "output": {
        "shape_name": "AssumeRoleResponse",
        "type": "structure",
        "members": {
          "Credentials": {
            "shape_name": "Credentials",
            "type": "structure",
            "members": {
              "AccessKeyId": {
                "shape_name": "accessKeyIdType",
                "type": "string",
                "min_length": 16,
                "max_length": 32,
                "pattern": "[\\w]*",
                "documentation": "docs",
                "required": True
              },
              "SecretAccessKey": {
                "shape_name": "accessKeySecretType",
                "type": "string",
                "documentation": "docs",
                "required": True
              },
              "SessionToken": {
                "shape_name": "tokenType",
                "type": "string",
                "documentation": "docs",
                "required": True
              },
              "Expiration": {
                "shape_name": "dateType",
                "type": "timestamp",
                "documentation": "docs",
                "required": True
              }
            },
            "documentation": "docs"
          },
          "AssumedRoleUser": {
            "shape_name": "AssumedRoleUser",
            "type": "structure",
            "members": {
              "AssumedRoleId": {
                "shape_name": "assumedRoleIdType",
                "type": "string",
                "min_length": 2,
                "max_length": 96,
                "pattern": "[\\w+=,.@:-]*",
                "documentation": "docs",
                "required": True
              },
              "Arn": {
                "shape_name": "arnType",
                "type": "string",
                "min_length": 20,
                "max_length": 2048,
                "documentation": "docs",
                "required": True
              }
            },
            "documentation": "docs"
          },
          "PackedPolicySize": {
            "shape_name": "nonNegativeIntegerType",
            "type": "integer",
            "min_length": 0,
            "documentation": "docs"
          },
          "NextToken": {
              "shape_name": "String",
              "type": "string",
              "documentation": None,
              "xmlname": "nextToken"
          },
        },
        "documentation": "docs"
      },
      "errors": [
        {
          "shape_name": "MalformedPolicyDocumentException",
          "type": "structure",
          "members": {
            "message": {
              "shape_name": "malformedPolicyDocumentMessage",
              "type": "string",
              "documentation": "docs"
            }
          },
          "documentation": "docs"
        },
        {
          "shape_name": "PackedPolicyTooLargeException",
          "type": "structure",
          "members": {
            "message": {
              "shape_name": "packedPolicyTooLargeMessage",
              "type": "string",
              "documentation": "docs"
            }
          },
          "documentation": "docs"
        }
      ],
      "documentation": "docs"
    },
    "RealOperation2013_02_04": {
      "name": "RealOperation2013_02_04",
      "input": {},
      "output": {
        "shape_name": "RealOperation2013_02_04Response",
        "type": "structure",
        "members": {
          "Result": {
            "shape_name": "Result",
            "type": "string",
            "documentation": ""
          }
        }
      },
      "errors": [],
      "documentation": "docs"
    },
    "NoOutputOperation": {
      "name": "NoOutputOperation",
      "input": {},
      "output": {},
      "errors": [],
      "documentation": "docs"
    },
    "DeprecatedOperation": {
      "input": {
        "shape_name": "DeprecatedOperationRequest",
        "type": "structure",
        "members": {
          "FooBar": {
            "shape_name": "foobarType",
            "type": "string",
            "documentation": "blah blah <![CDATA[\n\nfoobar ]]>blah blah",
          },
          "FieBaz": {
            "shape_name": "fiebazType",
            "type": "string",
            "documentation": "Don't use this, it's deprecated"
          }
        }
      },
      "documentation": "This is my <![CDATA[none of \nthis stuff should be here]]> stuff"
    },
    "DeprecatedOperation2": {
      "input": {
        "shape_name": "DeprecatedOperation2Request",
        "type": "structure",
        "members": {
          "FooBar": {
            "shape_name": "foobarType",
            "type": "string",
            "documentation": "blah blah blah blah",
          },
          "FieBaz": {
            "shape_name": "fiebazType",
            "type": "string",
            "documentation": ""
          }
        }
      },
      "documentation": "This operation has been deprecated."
    },
    "RenameOperation": {
      "input": {
        "shape_name": "RenameOperation",
        "type": "structure",
        "members": {
          "RenameMe": {
            "shape_name": "RenameMe",
            "type": "string",
            "documentation": "blah blah blah blah",
          },
          "FieBaz": {
            "shape_name": "fiebazType",
            "type": "string",
            "documentation": ""
          }
        }
      },
      "documentation": "This operation has been deprecated."
    },
    "EchoedOutputParams": {
      "name": "EchoedOutputParams",
      "input": {
        "shape_name": "EchoedOutputParams",
        "type": "structure",
        "members": {
          "Marker": {
            "shape_name": "String",
            "type": "string",
            "documentation": "blah blah blah blah",
          },
        }
      },
      "documentation": "",
      "output": {
        "shape_name": "EchoedOutputParamsResponse",
        "type": "structure",
        "members": {
          "Marker": {
            "shape_name": "String",
            "type": "string",
            "documentation": "",
          },
          "NextMarker": {
            "shape_name": "String",
            "type": "string",
            "documentation": "",
          },
          "ResultKey": {
            "shape_name": "String",
            "type": "string",
            "documentation": "",
          },
        }
      },
    }
  }
}


class TestTranslateExtensions(unittest.TestCase):
    def setUp(self):
        self.model = ModelFiles(SERVICES, {}, {}, {})

    def test_can_add_extras_top_level_keys(self):
        # A use case here would be adding the iterator/waiter configs.
        new_keys = {'extra': {'paginators': 'paginator_config'}}
        self.model.enhancements = new_keys
        new_model = translate(self.model)
        # There should be a new top level key 'iterators' that was merged in.
        self.assertEqual(new_model['paginators'], 'paginator_config')
        self.assertIn('operations', new_model)

    def test_can_add_fields_to_operation(self):
        # A use case would be to add checksum info for a param.

        # We could go for a more streamlined syntax, but this way, it's
        # clear how this maps onto the existing json model.
        new_keys = {
            'operations': {
                'AssumeRole': {
                    'input': {
                        'checksum': 'md5',
                    }
                },
            }
        }
        self.model.enhancements = new_keys
        new_model = translate(self.model)
        self.assertEqual(
            new_model['operations']['AssumeRole']['input']['checksum'],
            'md5')

    def test_can_add_fields_to_op_params(self):
        # A use case would be if we want to annotate that a
        # string type might also come from a file (keypairs, s3 uploads, etc).
        new_keys = {
            'operations': {
                'AssumeRole': {
                    'input': {
                        'members': {
                            'Policy': {
                                'alsofrom': 'filename',
                            },
                        }
                    }
                },
            }
        }
        self.model.enhancements = new_keys
        new_model = translate(self.model)
        self.assertEqual(
            new_model['operations']['AssumeRole']['input']['members']\
                    ['Policy']['alsofrom'],
            'filename')
        self.assertEqual(
            new_model['operations']['AssumeRole']['input']['members']\
                    ['Policy']['type'],
            'string')
        self.assertEqual(
            new_model['operations']['AssumeRole']['input']['members']\
                    ['RoleArn']['shape_name'],
            'arnType')


class TestTranslateModel(unittest.TestCase):
    def setUp(self):
        self.model = ModelFiles(SERVICES, {}, {}, {})

    def test_operations_is_a_dict_not_list(self):
        # In order to make overriding easier, we want the list of
        # operations to be a dict, not a list.  The way we don't have
        # to search through the list to find the operation we want to
        # change.  It also makes it easier to annontate operations.
        new_model = translate(self.model)
        self.assertIn('AssumeRole', new_model['operations'])

    def test_iterators_are_merged_into_operations(self):
        # This may or may not pan out, but if a pagination config is
        # specified, that info is merged into the specific operations.
        extra = {
            'pagination': {
                'AssumeRole': {
                    'input_token': 'NextToken',
                    'output_token': 'NextToken',
                    'limit_key': 'MaxResults',
                    'result_key': 'Credentials',
                    'non_aggregate_keys': ['PackedPolicySize',
                                           'AssumedRoleUser'],
                }
            }
        }
        self.model.enhancements = extra
        new_model = translate(self.model)
        op = new_model['operations']['AssumeRole']
        self.assertDictEqual(
            op['pagination'], {
                'input_token': 'NextToken',
                'py_input_token': 'next_token',
                'output_token': 'NextToken',
                'limit_key': 'MaxResults',
                'result_key': 'Credentials',
                'non_aggregate_keys': ['PackedPolicySize',
                                        'AssumedRoleUser'],
            })

    def test_py_input_name_is_not_added_if_it_exists(self):
        # This may or may not pan out, but if a pagination config is
        # specified, that info is merged into the specific operations.
        extra = {
            'pagination': {
                'AssumeRole': {
                    'input_token': 'NextToken',
                    'output_token': 'NextToken',
                    'py_input_token': 'other_value',
                    'limit_key': 'MaxResults',
                    'result_key': 'Credentials',
                    'non_aggregate_keys': ['PackedPolicySize',
                                           'AssumedRoleUser'],
                }
            }
        }
        self.model.enhancements = extra
        new_model = translate(self.model)
        op = new_model['operations']['AssumeRole']
        # Note how 'py_input_token' is left untouched.  This allows us
        # to override this naming if we need to.
        self.assertDictEqual(
            op['pagination'], {
                'input_token': 'NextToken',
                'py_input_token': 'other_value',
                'output_token': 'NextToken',
                'limit_key': 'MaxResults',
                'result_key': 'Credentials',
                    'non_aggregate_keys': ['PackedPolicySize',
                                           'AssumedRoleUser'],
            })

    def test_paginators_are_validated(self):
        # Can't create a paginator config that refers to a non existent
        # operation.
        extra = {
            'pagination': {
                'UnknownOperation': {
                    'input_token': 'NextToken',
                    'output_token': 'NextToken',
                    'max_results': 'MaxResults',
                }
            }
        }
        self.model.enhancements = extra
        with self.assertRaises(ValueError):
            new_model = translate(self.model)

    def test_paginators_are_placed_into_top_level_key(self):
        extra = {
            'pagination': {
                'AssumeRole': {
                    'input_token': 'NextToken',
                    'output_token': 'NextToken',
                    'result_key': 'Credentials',
                    'non_aggregate_keys': ['PackedPolicySize',
                                           'AssumedRoleUser'],
                }
            }
        }
        self.model.enhancements = extra
        new_model = translate(self.model)
        self.assertEqual(new_model['pagination'], extra['pagination'])

    def test_extra_key(self):
        # Anything in "extra" is merged as a top level key.
        extra = {
            "extra": {
                "signature_version": "v2",
            }
        }
        self.model.enhancements = extra
        new_model = translate(self.model)
        self.assertEqual(new_model['signature_version'], 'v2')
        self.assertEqual(new_model['documentation'], 'docs')

    def test_paginator_with_multiple_input_outputs(self):
        extra = {
            'pagination': {
                'AssumeRole': {
                    'input_token': ['NextToken', 'TokenToken'],
                    'output_token': ['NextToken'],
                    'result_key': 'Credentials',
                    'non_aggregate_keys': ['PackedPolicySize',
                                           'AssumedRoleUser'],
                }
            }
        }
        self.model.enhancements = extra
        new_model = translate(self.model)
        op = new_model['operations']['AssumeRole']
        self.assertDictEqual(
            op['pagination'], {
                'input_token': ['NextToken', 'TokenToken'],
                'py_input_token': ['next_token', 'token_token'],
                'output_token': ['NextToken'],
                'result_key': 'Credentials',
                'non_aggregate_keys': ['PackedPolicySize',
                                        'AssumedRoleUser'],
            })

    def test_result_key_validation(self):
        # result_key must exist.
        extra = {
            'pagination': {
                'AssumeRole': {
                    'input_token': ['Token', 'TokenToken'],
                    'output_token': ['NextToken']
                }
            }
        }
        self.model.enhancements = extra
        with self.assertRaises(ValueError):
            translate(self.model)

    def test_result_key_exists_in_output(self):
        extra = {
            'pagination': {
                'AssumeRole': {
                    'input_token': ['Token', 'TokenToken'],
                    'output_token': ['NextToken', 'NextTokenToken'],
                    'result_key': 'DoesNotExist',
                }
            }
        }
        self.model.enhancements = extra
        with self.assertRaises(ValueError):
            translate(self.model)

    def test_result_key_can_be_a_list(self):
        extra = {
            'pagination': {
                'AssumeRole': {
                    'input_token': ['NextToken'],
                    'output_token': ['NextToken'],
                    'result_key': ['Credentials', 'AssumedRoleUser'],
                    'non_aggregate_keys': ['PackedPolicySize'],
                }
            }
        }
        self.model.enhancements = extra
        new_model = translate(self.model)
        self.assertEqual(new_model['pagination'], extra['pagination'])

    def test_expected_schema_exists(self):
        # In this case, the key 'output_tokens' is suppose to be 'output_token'
        # so we should get an error when this happens.
        extra = {
            'pagination': {
                'AssumeRole': {
                    'input_token': ['Token', 'TokenToken'],
                    'output_tokens': ['NextToken', 'NextTokenToken'],
                    'result_key': ['Credentials', 'AssumedRoleUser'],
                }
            }
        }
        self.model.enhancements = extra
        with self.assertRaises(ValueError):
            translate(self.model)

    def test_input_tokens_exist_in_model(self):
        extra = {
            'pagination': {
                'AssumeRole': {
                    # In this case, "DoesNotExist" token is not in the input
                    # model, so we get an exception complaining about this.
                    'input_token': ['NextToken', 'DoesNotExist'],
                    'output_token': ['NextToken', 'NextTokenToken'],
                    'result_key': ['Credentials', 'AssumedRoleUser'],
                }
            }
        }
        self.model.enhancements = extra
        with self.assertRaises(ValueError):
            translate(self.model)

    def test_validate_limit_key_is_in_input(self):
        extra = {
            'pagination': {
                'AssumeRole': {
                    'input_token': 'NextToken',
                    'output_token': ['NextToken', 'NextTokenToken'],
                    'result_key': ['Credentials', 'AssumedRoleUser'],
                    # In this case, "DoesNotExist" token is not in the input
                    # model, so we get an exception complaining about this.
                    'limit_key': 'DoesNotExist',
                }
            }
        }
        self.model.enhancements = extra
        with self.assertRaises(ValueError):
            translate(self.model)

    def test_cant_add_pagination_to_nonexistent_operation(self):
        extra = {
            'pagination': {
                'ThisOperationDoesNotExist': {
                    'input_token': 'NextToken',
                    'output_token': ['NextToken', 'NextTokenToken'],
                    'result_key': ['Credentials', 'AssumedRoleUser'],
                    'limit_key': 'Foo',
                }
            }
        }
        self.model.enhancements = extra
        with self.assertRaisesRegexp(
                ValueError, "Trying to add pagination config for non "
                            "existent operation: ThisOperationDoesNotExist"):
            translate(self.model)

    def test_skip_jmespath_validation(self):
        # This would fail previously.
        extra = {
            'pagination': {
                'AssumeRole': {
                    'input_token': ['NextToken'],
                    'output_token': ['NextToken'],
                    'result_key': 'Credentials.AssumedRoleUser',
                }
            }
        }
        self.model.enhancements = extra
        new_model = translate(self.model)
        self.assertEqual(new_model['pagination'], extra['pagination'])

    def test_result_key_validation_with_no_output(self):
        extra = {
            'pagination': {
                # RealOperation does not have any output members so
                # we should get an error message telling us this.
                'NoOutputOperation': {
                    'input_token': 'NextToken',
                    'output_token': ['NextToken', 'NextTokenToken'],
                    'result_key': ['Credentials', 'AssumedRoleUser'],
                    'limit_key': 'Foo',
                }
            }
        }
        self.model.enhancements = extra
        with self.assertRaisesRegexp(
                ValueError, "Trying to add pagination config for an "
                            "operation with no output members: "
                            "NoOutputOperation"):
            translate(self.model)

    def test_echoed_input_params_ignored(self):
        extra = {
            'pagination': {
                'EchoedOutputParams': {
                    'input_token': ['Marker'],
                    'output_token': ['NextMarker'],
                    'result_key': 'ResultKey',
                }
            }
        }
        self.model.enhancements = extra
        new_model = translate(self.model)
        self.assertEqual(new_model['pagination'], extra['pagination'])


class TestBuildRetryConfig(unittest.TestCase):
    def setUp(self):
        self.retry = {
            "definitions": {
                "def_name": {
                    "from": {"definition": "file"}
                }
            },
            "retry": {
                "__default__": {
                    "max_attempts": 5,
                    "delay": "global_delay",
                    "policies": {
                        "global_one": "global",
                        "override_me": "global",
                    }
                },
                "sts": {
                    "__default__": {
                        "delay": "service_specific_delay",
                        "policies": {
                            "service_one": "service",
                            "override_me": "service",
                        }
                    },
                    "AssumeRole": {
                        "policies": {
                            "name": "policy",
                            "other": {"$ref": "def_name"}
                        }
                    }
                }
            }
        }

    def test_inject_retry_config(self):
        model = ModelFiles(SERVICES, self.retry, {})
        new_model = translate(model)
        self.assertIn('retry', new_model)
        retry = new_model['retry']
        self.assertIn('__default__', retry)
        self.assertEqual(
            retry['__default__'], {
                "max_attempts": 5,
                "delay": "service_specific_delay",
                "policies": {
                    "global_one": "global",
                    "override_me": "service",
                    "service_one": "service",
                }
            }
        )
        # Policies should be merged.
        operation_config = retry['AssumeRole']
        self.assertEqual(operation_config['policies']['name'], 'policy')

    def test_resolve_reference(self):
        model = ModelFiles(SERVICES, self.retry, {})
        new_model = translate(model)
        operation_config = new_model['retry']['AssumeRole']
        # And we should resolve references.
        self.assertEqual(operation_config['policies']['other'],
                         {"from": {"definition": "file"}})


class TestReplacePartOfOperation(unittest.TestCase):
    def test_replace_operation_key_name(self):
        enhancements = {
            'transformations': {
                'operation-name': {'remove': r'\d{4}_\d{2}_\d{2}'}
            }
        }
        model = ModelFiles(SERVICES, retry={},
                           enhancements=enhancements)
        new_model = translate(model)
        # But the key into the operation dict is stripped of the
        # matched regex.
        self.assertEqual(list(sorted(new_model['operations'].keys())),
                         ['AssumeRole', 'DeprecatedOperation',
                          'DeprecatedOperation2', 'EchoedOutputParams',
                          'NoOutputOperation','RealOperation',
                          'RenameOperation'])
        # But the name key attribute is left unchanged.
        self.assertEqual(new_model['operations']['RealOperation']['name'],
                         'RealOperation2013_02_04')

    def test_merging_occurs_after_transformation(self):
        enhancements = {
            'transformations': {
                'operation-name': {'remove': r'\d{4}_\d{2}_\d{2}'}
            },
            'operations': {
                'RealOperation': {
                    'input': {
                        'checksum': 'md5',
                    }
                },
            }
        }
        model = ModelFiles(SERVICES, retry={}, enhancements=enhancements)
        model.enhancements = enhancements
        new_model = translate(model)
        self.assertIn('RealOperation', new_model['operations'])
        self.assertEqual(
            new_model['operations']['RealOperation']['input']['checksum'],
            'md5')


class TestRemovalOfDeprecatedParams(unittest.TestCase):

    def test_remove_deprecated_params(self):
        enhancements = {
            'transformations': {
                'remove-deprecated-params': {'deprecated_keyword': 'deprecated'}
                }
            }
        model = ModelFiles(SERVICES, retry={}, enhancements=enhancements)
        new_model = translate(model)
        operation = new_model['operations']['DeprecatedOperation']
        # The deprecated param should be gone, the other should remain
        self.assertIn('FooBar', operation['input']['members'])
        self.assertNotIn('FieBaz', operation['input']['members'])

class TestRemovalOfDeprecatedOps(unittest.TestCase):

    def test_remove_deprecated_ops(self):
        enhancements = {
            'transformations': {
                'remove-deprecated-operations':
                    {'deprecated_keyword': 'deprecated'}
                }
            }
        model = ModelFiles(SERVICES, retry={}, enhancements=enhancements)
        new_model = translate(model)
        # The deprecated operation should be gone
        self.assertNotIn('DeprecatedOperation2', new_model['operations'])


class TestFilteringOfDocumentation(unittest.TestCase):

    def test_remove_deprecated_params(self):
        enhancements = {
            "transformations": {
                "filter-documentation": {
                    "filter": {
                        "regex": "<!\\[CDATA\\[.*\\]\\]>",
                        "replacement": ""
                        }
                    }
                }
            }
        model = ModelFiles(SERVICES, retry={}, enhancements=enhancements)
        new_model = translate(model)
        operation = new_model['operations']['DeprecatedOperation']
        # The deprecated param should be gone, the other should remain
        self.assertEqual(operation['documentation'], 'This is my  stuff')
        param = operation['input']['members']['FooBar']
        self.assertEqual(param['documentation'], 'blah blah blah blah')


class TestRenameParams(unittest.TestCase):
    def test_rename_param(self):
        enhancements = {
            'transformations': {
                'renames': {
                    'RenameOperation.input.members.RenameMe': 'BeenRenamed',
                }
            }
        }
        model = ModelFiles(SERVICES, retry={}, enhancements=enhancements)
        new_model = translate(model)
        arguments = new_model['operations']['RenameOperation']['input']['members']
        self.assertNotIn('RenameMe', arguments)
        self.assertIn('BeenRenamed', arguments)


class TestWaiterDenormalization(unittest.TestCase):
    maxDiff = None

    def setUp(self):
        self.model = ModelFiles(SERVICES, {}, {})

    def test_waiter_default_resolved(self):
        extra = {
            'waiters': {
                '__default__': {
                    'interval': 20,
                    'operation': 'AssumeRole',
                    'max_attempts': 25,
                    'acceptor_type': 'output',
                    'acceptor_path': 'path',
                    'acceptor_value': 'value',
                },
                # Note that this config doesn't make any actual sense,
                # this is just testing we denormalize fields properly.
                'RoleExists': {
                    'operation': 'AssumeRole',
                    'ignore_errors': ['Error1'],
                    'success_type': 'output',
                    'success_path': 'Table.TableStatus',
                    'success_value': ['ACTIVE'],
                }
            }
        }
        self.model.enhancements = extra
        new_model = translate(self.model)
        denormalized = {
            'RoleExists': {
                'interval': 20,
                'max_attempts': 25,
                'operation': 'AssumeRole',
                'ignore_errors': ['Error1'],
                'success': {
                    'type': 'output',
                    'path': 'Table.TableStatus',
                    'value': ['ACTIVE'],
                },
                'failure': {
                    'type': 'output',
                    'path': 'path',
                    'value': ['value'],
                }
            }
        }
        self.assertEqual(new_model['waiters'], denormalized)

    def test_default_and_extends(self):
        extra = {
            'waiters': {
                '__default__': {
                    'interval': 20,
                    'max_attempts': 25,
                },
                '__RoleResource': {
                    'operation': 'AssumeRole',
                    'max_attempts': 50,
                },
                'RoleExists': {
                    'extends': '__RoleResource',
                    'ignore_errors': ['Error1'],
                    'success_type': 'output',
                    'success_path': 'Table.TableStatus',
                    'success_value': 'ACTIVE',
                    'failure_type': 'output',
                    'failure_path': 'Table.TableStatus',
                    'failure_value': 'ACTIVE',
                }
            }
        }
        self.model.enhancements = extra
        new_model = translate(self.model)
        denormalized = {
            'RoleExists': {
                # From __default__
                'interval': 20,
                # Overriden from __RoleResource
                'max_attempts': 50,
                # Defined in __RoleResource
                'operation': 'AssumeRole',
                # Defined in RoleExists
                'ignore_errors': ['Error1'],
                'success': {
                    'type': 'output',
                    'path': 'Table.TableStatus',
                    'value': ['ACTIVE'],
                },
                'failure': {
                    'type': 'output',
                    'path': 'Table.TableStatus',
                    'value': ['ACTIVE'],
                }
            }
        }
        self.assertEqual(new_model['waiters'], denormalized)

    def test_acceptor_path_resolution(self):
        extra = {
            'waiters': {
                'RoleExists': {
                    'operation': 'AssumeRole',
                    'acceptor_type': 'output',
                    'acceptor_path': 'acceptor_path',
                    'success_value': 'success_value',
                    'failure_value': 'failure_value',
                }
            }
        }
        self.model.enhancements = extra
        new_model = translate(self.model)
        denormalized = {
            'RoleExists': {
                'operation': 'AssumeRole',
                # We should only have success/failure values,
                # no acceptor types, those are all resolved.
                'success': {
                    # From acceptor_type.
                    'type': 'output',
                    # From acceptor_path.
                    'path': 'acceptor_path',
                    # From success_value.
                    'value': ['success_value'],
                },
                'failure': {
                    # From acceptor_type.
                    'type': 'output',
                    # From acceptor_path.
                    'path': 'acceptor_path',
                    # From failure_value.
                    'value': ['failure_value'],
                }
            }
        }
        self.assertEqual(new_model['waiters'], denormalized)

    def test_only_acceptors_provided(self):
        extra = {
            'waiters': {
                'RoleExists': {
                    'operation': 'AssumeRole',
                    'acceptor_type': 'output',
                    'acceptor_path': 'acceptor_path',
                    'acceptor_value': 'acceptor_value',
                }
            }
        }
        self.model.enhancements = extra
        new_model = translate(self.model)
        denormalized = {
            'RoleExists': {
                'operation': 'AssumeRole',
                'success': {
                    'type': 'output',
                    'path': 'acceptor_path',
                    'value': ['acceptor_value'],
                },
                'failure': {
                    'type': 'output',
                    'path': 'acceptor_path',
                    'value': ['acceptor_value'],
                }
            }
        }
        self.assertEqual(new_model['waiters'], denormalized)

    def test_validate_valid_operation(self):
        extra = {
            'waiters': {
                '__default__': {
                    'interval': 20,
                    'max_attempts': 25,
                },
                '__RoleResource': {
                    'operation': 'THISOPERATIONDOESNOTEXIST',
                    'max_attempts': 50,
                },
                'RoleExists': {
                    'extends': '__RoleResource',
                    'ignore_errors': ['Error1'],
                    'success_type': 'output',
                    'success_path': 'Table.TableStatus',
                    'success_value': 'ACTIVE',
                    'failure_type': 'output',
                    'failure_path': 'Table.TableStatus',
                    'failure_value': 'ACTIVE',
                }
            }
        }
        self.model.enhancements = extra
        with self.assertRaises(ValueError):
            new_model = translate(self.model)

    def test_only_success_defined(self):
        extra = {
            'waiters': {
                '__default__': {
                    "interval": 15,
                    "max_attempts": 40,
                    "acceptor_type": "output"
                },
                'AssumeRole': {
                    "operation": "AssumeRole",
                    "success_path": "Snapshots[].State",
                    "success_value": "completed"
                }
            }
        }
        self.model.enhancements = extra
        new_model = translate(self.model)
        denormalized = {
            'AssumeRole': {
                'interval': 15,
                'max_attempts': 40,
                'operation': 'AssumeRole',
                'success': {
                    'type': 'output',
                    'path': 'Snapshots[].State',
                    'value': ['completed'],
                },
                # This is technically "incorrect", in the sense that this is
                # not a valid waiter config, but the waiter module has tests
                # to verify that it handles this "incorrect" case.
                'failure': {
                    'type': 'output',
                }
            }
        }
        self.assertEqual(new_model['waiters'], denormalized)


class TestResemblesJMESPath(unittest.TestCase):
    maxDiff = None

    def test_is_jmespath(self):
      self.assertTrue(resembles_jmespath_exp('Something.Else'))
      self.assertTrue(resembles_jmespath_exp('Something[1]'))

    def test_is_not_jmespath(self):
      self.assertFalse(resembles_jmespath_exp('Something'))


if __name__ == '__main__':
    unittest.main()



# Copyright (c) 2012-2013 Mitch Garnaat http://garnaat.org/
# Copyright 2012-2014 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.

"""Translate the raw json files into python specific descriptions."""
import os
import re
from copy import deepcopy

import jmespath

from botocore.compat import OrderedDict, json
from botocore.utils import merge_dicts
from botocore import xform_name


class ModelFiles(object):
    """Container object to hold all the various parsed json files.

    Includes:

        * The json service description.
        * The _retry.json file.
        * The <service>.extra.json enhancements file.
        * The name of the service.

    """
    def __init__(self, model, retry, enhancements, name=''):
        self.model = model
        self.retry = retry
        self.enhancements = enhancements
        self.name = name


def load_model_files(args):
    model = json.load(open(args.modelfile),
                      object_pairs_hook=OrderedDict)
    retry = json.load(open(args.retry_file),
                      object_pairs_hook=OrderedDict)
    enhancements = _load_enhancements_file(args.enhancements_file)
    service_name = os.path.splitext(os.path.basename(args.modelfile))[0]
    return ModelFiles(model, retry, enhancements, name=service_name)


def _load_enhancements_file(file_path):
    if not os.path.isfile(file_path):
        return {}
    else:
        return json.load(open(file_path),
                         object_pairs_hook=OrderedDict)


def translate(model):
    new_model = deepcopy(model.model)
    new_model.update(model.enhancements.get('extra', {}))
    try:
        del new_model['pagination']
    except KeyError:
        pass
    handle_op_renames(new_model, model.enhancements)
    handle_remove_deprecated_params(new_model, model.enhancements)
    handle_remove_deprecated_operations(new_model, model.enhancements)
    handle_filter_documentation(new_model, model.enhancements)
    handle_rename_params(new_model, model.enhancements)
    add_pagination_configs(
        new_model,
        model.enhancements.get('pagination', {}))
    add_waiter_configs(
        new_model,
        model.enhancements.get('waiters', {}))
    # Merge in any per operation overrides defined in the .extras.json file.
    merge_dicts(new_model['operations'],
                model.enhancements.get('operations', {}))
    add_retry_configs(
        new_model, model.retry.get('retry', {}),
        definitions=model.retry.get('definitions', {}))
    return new_model


def handle_op_renames(new_model, enhancements):
    # This allows for operations to be renamed.  The only
    # implemented transformation is removing part of the operation name
    # (because that's all we currently need.)
    remove = enhancements.get('transformations', {}).get(
        'operation-name', {}).get('remove')
    if remove is not None:
        # We're going to recreate the dictionary because we want to preserve
        # the order.  This is the only option we have unless we have our own
        # custom OrderedDict.
        remove_regex = re.compile(remove)
        operations = new_model['operations']
        new_operation = OrderedDict()
        for key in operations:
            new_key = remove_regex.sub('', key)
            new_operation[new_key] = operations[key]
        new_model['operations'] = new_operation


def handle_remove_deprecated_operations(new_model, enhancements):
    # This removes any operation whose documentation string contains
    # the specified phrase that marks a deprecated parameter.
    keyword = enhancements.get('transformations', {}).get(
        'remove-deprecated-operations', {}).get('deprecated_keyword')
    remove = []
    if keyword is not None:
        operations = new_model['operations']
        for op_name in operations:
            operation = operations[op_name]
            if operation:
                docs = operation['documentation']
                if docs and docs.find(keyword) >= 0:
                    remove.append(op_name)
    for op in remove:
        del new_model['operations'][op]


def handle_remove_deprecated_params(new_model, enhancements):
    # This removes any parameter whose documentation string contains
    # the specified phrase that marks a deprecated parameter.
    keyword = enhancements.get('transformations', {}).get(
        'remove-deprecated-params', {}).get('deprecated_keyword')
    if keyword is not None:
        operations = new_model['operations']
        for op_name in operations:
            operation = operations[op_name]
            params = operation.get('input', {}).get('members')
            if params:
                new_params = OrderedDict()
                for param_name in params:
                    param = params[param_name]
                    docs = param['documentation']
                    if docs and docs.find(keyword) >= 0:
                        continue
                    new_params[param_name] = param
                operation['input']['members'] = new_params


def _filter_param_doc(param, replacement, regex):
    # Recurse into complex parameters looking for documentation.
    doc = param.get('documentation')
    if doc:
        param['documentation'] = regex.sub(replacement, doc)
    if param['type'] == 'structure':
        for member_name in param['members']:
            member = param['members'][member_name]
            _filter_param_doc(member, replacement, regex)
    if param['type'] == 'map':
        _filter_param_doc(param['keys'], replacement, regex)
        _filter_param_doc(param['members'], replacement, regex)
    elif param['type'] == 'list':
        _filter_param_doc(param['members'], replacement, regex)


def handle_filter_documentation(new_model, enhancements):
    # This provides a way to filter undesireable content (e.g. CDATA)
    # from documentation strings.
    doc_filter = enhancements.get('transformations', {}).get(
        'filter-documentation', {}).get('filter')
    if doc_filter is not None:
        filter_regex = re.compile(doc_filter.get('regex', ''), re.DOTALL)
        replacement = doc_filter.get('replacement')
        operations = new_model['operations']
        for op_name in operations:
            operation = operations[op_name]
            doc = operation.get('documentation')
            if doc:
                new_doc = filter_regex.sub(replacement, doc)
                operation['documentation'] = new_doc
            params = operation.get('input', {}).get('members')
            if params:
                for param_name in params:
                    param = params[param_name]
                    _filter_param_doc(param, replacement, filter_regex)


def handle_rename_params(new_model, enhancements):
    renames = enhancements.get('transformations', {}).get(
        'renames', {})
    if not renames:
        return
    # This is *extremely* specific to botocore's translations, but
    # we support a restricted set of argument renames based on a
    # jmespath expression.
    for expression, new_value in renames.items():
        # First we take everything up until the last dot.
        parent_expression, key = expression.rsplit('.', 1)
        matched = jmespath.search(parent_expression, new_model['operations'])
        current = matched[key]
        del matched[key]
        matched[new_value] = current


def resembles_jmespath_exp(value):
    # For now, we'll do a naive check.
    if '.' in value or '[' in value:
        return True
    return False


def add_pagination_configs(new_model, pagination):
    # Adding in pagination configs means copying the config to a top level
    # 'pagination' key in the new model, and it also means adding the
    # pagination config to each individual operation.
    # Also, the input_token needs to be transformed to the python specific
    # name, so we're adding a py_input_token (e.g. NextToken -> next_token).
    if pagination:
        new_model['pagination'] = pagination
    for name in pagination:
        config = pagination[name]
        _check_known_pagination_keys(config)
        if 'py_input_token' not in config:
            _add_py_input_token(config)
        _validate_result_key_exists(config)
        _validate_referenced_operation_exists(new_model, name)
        operation = new_model['operations'][name]
        _validate_operation_has_output(operation, name)
        _check_input_keys_match(config, operation)
        _check_output_keys_match(config, operation,
                                 new_model.get('endpoint_prefix', ''))
        operation['pagination'] = config.copy()


def _validate_operation_has_output(operation, name):
    if not operation['output']:
        raise ValueError("Trying to add pagination config for an "
                         "operation with no output members: %s" % name)


def _validate_referenced_operation_exists(new_model, name):
    if name not in new_model['operations']:
        raise ValueError("Trying to add pagination config for non "
                         "existent operation: %s" % name)


def _validate_result_key_exists(config):
    # result_key must be defined.
    if 'result_key' not in config:
        raise ValueError("Required key 'result_key' is missing from "
                         "from pagination config: %s" % config)


def _add_py_input_token(config):
    input_token = config['input_token']
    if isinstance(input_token, list):
        py_input_token = []
        for token in input_token:
            py_input_token.append(xform_name(token))
        config['py_input_token'] = py_input_token
    else:
        config['py_input_token'] = xform_name(input_token)


def add_waiter_configs(new_model, waiters):
    if waiters:
        denormalized = denormalize_waiters(waiters)
        # Before adding it to the new model, we need to verify the
        # final denormalized model.
        for value in denormalized.values():
            if value['operation'] not in new_model['operations']:
                raise ValueError()
        new_model['waiters'] = denormalized


def denormalize_waiters(waiters):
    # The waiter configuration is normalized to avoid duplication.
    # You can inherit defaults, and extend from other definitions.
    # We're going to denormalize this so that the implementation for
    # consuming waiters is simple.
    default = waiters.get('__default__', {})
    new_waiters = {}
    for key, value in waiters.items():
        if key.startswith('__'):
            # Keys that start with '__' are considered abstract/internal
            # and are only used for inheritance.  Because we're going
            # to denormalize the configs and perform all the lookups
            # during this translation process, the abstract/internal
            # configs don't need to make it into the final translated
            # config so we can just skip these.
            continue
        new_waiters[key] = denormalize_single_waiter(value, default, waiters)
    return new_waiters


def denormalize_single_waiter(value, default, waiters):
    """Denormalize a single waiter config.

    :param value: The dictionary of a single waiter config, e.g.
        the ``InstanceRunning`` or ``TableExists`` config.  This
        is the config we're going to denormalize.
    :param default: The ``__default__`` (if any) configuration.
        This is needed to resolve the lookup process.
    :param waiters: The full configuration of the waiters.
        This is needed if we need to look up at parent class that the
        current config extends.
    :return: The denormalized config.
    :rtype: dict

    """
    # First we need to resolve all the keys based on the inheritance
    # hierarchy.  The lookup process is:
    # The most bottom/leaf class is ``value``.  From there we need
    # to look up anything it inherits from (denoted via the ``extends``
    # key).  We need to perform this process recursively until we hit
    # a config that has no ``extends`` key.
    # And finally if we haven't found our value yet, we check in the
    # ``__default__`` key.
    # So the first thing we need to do is build the lookup chain that
    # starts with ``value`` and ends with ``__default__``.
    lookup_chain = [value]
    current = value
    while True:
        if 'extends' not in current:
            break
        current = waiters[current.get('extends')]
        lookup_chain.append(current)
    lookup_chain.append(default)
    new_waiter = {}
    # Now that we have this lookup chain we can build the entire set
    # of values by starting at the most parent class and walking down
    # to the children.  At each step the child is merged onto the parent's
    # config items.  This is the desired behavior as a child's values
    # overrides its parents.  This is what the ``reversed(...)`` call
    # is for.
    for element in reversed(lookup_chain):
        new_waiter.update(element)
    # We don't care about 'extends' so we can safely remove that key.
    new_waiter.pop('extends', {})
    # Now we need to resolve the success/failure values.  We
    # want to completely remove the acceptor types.
    # The logic here is that if there is no success/failure_* variable
    # defined, it inherits this value from the matching acceptor_* variable.
    new_waiter['success_type'] = new_waiter.get(
        'success_type', new_waiter.get('acceptor_type'))
    new_waiter['success_path'] = new_waiter.get(
        'success_path', new_waiter.get('acceptor_path'))
    new_waiter['success_value'] = new_waiter.get(
        'success_value', new_waiter.get('acceptor_value'))
    new_waiter['failure_type'] = new_waiter.get(
        'failure_type', new_waiter.get('acceptor_type'))
    new_waiter['failure_path'] = new_waiter.get(
        'failure_path', new_waiter.get('acceptor_path'))
    new_waiter['failure_value'] = new_waiter.get(
        'failure_value', new_waiter.get('acceptor_value'))
    # We can remove acceptor_* vars because they're only used for lookups
    # and we've already performed this step in the lines above.
    new_waiter.pop('acceptor_type', '')
    new_waiter.pop('acceptor_path', '')
    new_waiter.pop('acceptor_value', '')
    # Remove any keys with a None value.
    for key in list(new_waiter.keys()):
        if new_waiter[key] is None:
            del new_waiter[key]
    # Check required keys.
    for required in ['operation', 'success_type']:
        if required not in new_waiter:
            raise ValueError('Missing required waiter configuration '
                             'value "%s": %s' % (required, new_waiter))
        if new_waiter.get(required) is None:
            raise ValueError('Required waiter configuration '
                             'value cannot be None "%s": %s' %
                             (required, new_waiter))
    # Finally, success/failure values can be a scalar or a list.  We're going
    # to just always make them a list.
    if 'success_value' in new_waiter and not \
            isinstance(new_waiter['success_value'], list):
        new_waiter['success_value'] = [new_waiter['success_value']]
    if 'failure_value' in new_waiter and not \
            isinstance(new_waiter['failure_value'], list):
        new_waiter['failure_value'] = [new_waiter['failure_value']]
    _transform_waiter(new_waiter)
    return new_waiter


def _transform_waiter(new_waiter):
    # This transforms the waiters into a format that's slightly
    # easier to consume.
    if 'success_type' in new_waiter:
        success = {'type': new_waiter.pop('success_type')}
        if 'success_path' in new_waiter:
            success['path'] = new_waiter.pop('success_path')
        if 'success_value' in new_waiter:
            success['value'] = new_waiter.pop('success_value')
        new_waiter['success'] = success
    if 'failure_type' in new_waiter:
        failure = {'type': new_waiter.pop('failure_type')}
        if 'failure_path' in new_waiter:
            failure['path'] = new_waiter.pop('failure_path')
        if 'failure_value' in new_waiter:
            failure['value'] = new_waiter.pop('failure_value')
        new_waiter['failure'] = failure


def _check_known_pagination_keys(config):
    # Verify that the pagination config only has keys we expect to see.
    expected = set(['input_token', 'py_input_token', 'output_token',
                    'result_key', 'limit_key', 'more_results',
                    'non_aggregate_keys'])
    for key in config:
        if key not in expected:
            raise ValueError("Unknown key in pagination config: %s" % key)


def _check_output_keys_match(config, operation, service_name):
    output_members = list(operation['output']['members'])
    jmespath_seen = False
    for output_key in _get_all_page_output_keys(config):
        if resembles_jmespath_exp(output_key):
            # We don't validate jmespath expressions for now.
            jmespath_seen = True
            continue
        if output_key not in output_members:
            raise ValueError("Key %r is not an output member: %s" %
                             (output_key,
                              output_members))
        output_members.remove(output_key)
    # Some services echo the input parameters in the response
    # output.  We should not trigger a validation error
    # if those params are still not accounted for.
    for input_name in operation['input']['members']:
        if input_name in output_members:
            output_members.remove(input_name)
    if not jmespath_seen and output_members:
        # Because we can't validate jmespath expressions yet,
        # we can't say for user if output_members actually has
        # remaining keys or not.
        if service_name == 's3' and output_members == ['Name']:
            # The S3 model uses 'Name' for the output key, which
            # actually maps to the 'Bucket' input param so we don't
            # need to validate this output member.  This is the only
            # model that has this, so we can just special case this
            # for now.
            return
        raise ValueError("Output members still exist for operation %s: %s" % (
            operation['name'], output_members))


def _get_all_page_output_keys(config):
    if not isinstance(config['result_key'], list):
        yield config['result_key']
    else:
        for result_key in config['result_key']:
            yield result_key
    if not isinstance(config['output_token'], list):
        yield config['output_token']
    else:
        for result_key in config['output_token']:
            yield result_key
    if 'more_results' in config:
        yield config['more_results']
    for key in config.get('non_aggregate_keys', []):
        yield key


def _check_input_keys_match(config, operation):
    input_tokens = config['input_token']
    if not isinstance(input_tokens, list):
        input_tokens = [input_tokens]
    valid_input_names = operation['input']['members']
    for token in input_tokens:
        if token not in valid_input_names:
            raise ValueError("input_token refers to a non existent "
                             "input name for operation %s: %s.  "
                             "Must be one of: %s" % (operation['name'], token,
                                                     list(valid_input_names)))
    if 'limit_key' in config and config['limit_key'] not in valid_input_names:
        raise ValueError("limit_key refers to a non existent input name for "
                         "operation %s: %s.  Must be one of: %s" % (
                             operation['name'], config['limit_key'],
                             list(valid_input_names)))


def add_retry_configs(new_model, retry_model, definitions):
    if not retry_model:
        new_model['retry'] = {}
        return
    # The service specific retry config is keyed off of the endpoint
    # prefix as defined in the JSON model.
    endpoint_prefix = new_model.get('endpoint_prefix', '')
    final_retry_config = build_retry_config(endpoint_prefix, retry_model,
                                            definitions)
    new_model['retry'] = final_retry_config


def build_retry_config(endpoint_prefix, retry_model, definitions):
    service_config = retry_model.get(endpoint_prefix, {})
    resolve_references(service_config, definitions)
    # We want to merge the global defaults with the service specific
    # defaults, with the service specific defaults taking precedence.
    # So we use the global defaults as the base.
    final_retry_config = {'__default__': retry_model.get('__default__', {})}
    resolve_references(final_retry_config, definitions)
    # The merge the service specific config on top.
    merge_dicts(final_retry_config, service_config)
    return final_retry_config


def resolve_references(config, definitions):
    """Recursively replace $ref keys.

    To cut down on duplication, common definitions can be declared
    (and passed in via the ``definitions`` attribute) and then
    references as {"$ref": "name"}, when this happens the reference
    dict is placed with the value from the ``definition`` dict.

    This is recursively done.

    """
    for key, value in config.items():
        if isinstance(value, dict):
            if len(value) == 1 and list(value.keys())[0] == '$ref':
                # Then we need to resolve this reference.
                config[key] = definitions[list(value.values())[0]]
            else:
                resolve_references(value, definitions)



# Copyright 2015 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
# http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
from nose.tools import assert_equals

import botocore.session


def test_lint_waiter_configs():
    session = botocore.session.get_session()
    for service_name in session.get_available_services():
        client = session.create_client(service_name, 'us-east-1')
        service_model = client.meta.service_model
        for waiter_name in client.waiter_names:
            waiter = client.get_waiter(waiter_name)
            yield _lint_single_waiter, waiter, service_model


def _lint_single_waiter(waiter, service_model):
    operation_name = waiter.config.operation
    # Needs to reference an existing operation name.
    if operation_name not in service_model.operation_names:
        raise AssertionError("Waiter config references unknown "
                             "operation: %s" % operation_name)
    # Needs to have at least one acceptor.
    if not waiter.config.acceptors:
        raise AssertionError("Waiter config must have at least "
                             "one acceptor state: %s" % waiter.name)
    # Additional things to add:
    # 1. Verify the error acceptors correspond to a 'code' in the model
    # 2. Verify JMESPath expressions can resolve to something in
    #    the response.

